from PyQt5 import QtWidgets
from PyQt5 import QtCore
from PyQt5 import QtGui

from basedocgraphicsview import BaseDocGraphicsView

class DocGraphicsView(BaseDocGraphicsView):
    def __init__(self, parent, render_num=4):
        super(DocGraphicsView, self).__init__(parent)

        self.scene.setBackgroundBrush(QtGui.QBrush(QtCore.Qt.white)) # set background

    def wheelEvent(self, ev):
        # debug("wheelEvent in BaseDocGraphicsView")
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        if modifiers == QtCore.Qt.ControlModifier:
            delta = ev.angleDelta()
            if delta.y() > 0:
                self.zoomIn()
            else:
                self.zoomOut()
            ev.accept() # accept an event in order to stop it from propagating further
        else:
            return super().wheelEvent(ev) # call parent's handler, making the view scrolled (touch included)

    def mouseDoubleClickEvent(self, ev):
        # debug('mouseDoubleClickEvent in BaseDocGraphicsView')
        if self.current_zoom_index == 6 and self.fitwidth_flag == False:
            self.zoomFitWidth()
        else:
            # Zoom to 100%
            self.current_zoom_index = 6 # magic number, TODO
            assert(self.zoom_levels[self.current_zoom_index] == 1.0)
            self.fitwidth_flag = False
            self.redrawPages()
            self.zoomRatioChanged.emit(self.zoom_levels[self.current_zoom_index])
        return super().mouseDoubleClickEvent(ev)




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './kuafu/resources/ui_library.ui'
#
# Created by: PyQt5 UI code generator 5.15.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_librarywidget(object):
    def setupUi(self, librarywidget):
        librarywidget.setObjectName("librarywidget")
        librarywidget.resize(970, 701)
        self.horizontalLayout = QtWidgets.QHBoxLayout(librarywidget)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.splitter = QtWidgets.QSplitter(librarywidget)
        self.splitter.setMinimumSize(QtCore.QSize(0, 0))
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setChildrenCollapsible(False)
        self.splitter.setObjectName("splitter")
        self.tabWidget_2 = QtWidgets.QTabWidget(self.splitter)
        self.tabWidget_2.setMinimumSize(QtCore.QSize(200, 0))
        self.tabWidget_2.setMaximumSize(QtCore.QSize(600, 16777215))
        self.tabWidget_2.setObjectName("tabWidget_2")
        self.tab_3 = QtWidgets.QWidget()
        self.tab_3.setObjectName("tab_3")
        self.horizontal_Layout = QtWidgets.QHBoxLayout(self.tab_3)
        self.horizontal_Layout.setContentsMargins(0, 0, 0, 0)
        self.horizontal_Layout.setObjectName("horizontal_Layout")
        self.splitter_1 = QtWidgets.QSplitter(self.tab_3)
        self.splitter_1.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.splitter_1.setOrientation(QtCore.Qt.Vertical)
        self.splitter_1.setObjectName("splitter_1")
        self.fileview = QtWidgets.QTreeView(self.splitter_1)
        self.fileview.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.fileview.setObjectName("fileview")
        self.annotview = QtWidgets.QTreeView(self.splitter_1)
        self.annotview.setObjectName("annotview")
        self.horizontal_Layout.addWidget(self.splitter_1)
        self.tabWidget_2.addTab(self.tab_3, "")
        self.tab_4 = QtWidgets.QWidget()
        self.tab_4.setObjectName("tab_4")
        self.tabWidget_2.addTab(self.tab_4, "")
        self.splitter_doc = QtWidgets.QSplitter(self.splitter)
        self.splitter_doc.setMinimumSize(QtCore.QSize(600, 0))
        self.splitter_doc.setOrientation(QtCore.Qt.Vertical)
        self.splitter_doc.setObjectName("splitter_doc")
        self.preview_graphicsview = DocGraphicsView(self.splitter_doc)
        self.preview_graphicsview.setObjectName("preview_graphicsview")
        self.preview_graphicsview_2 = DocGraphicsView(self.splitter_doc)
        self.preview_graphicsview_2.setObjectName("preview_graphicsview_2")
        self.thumb_graphicsview = ThumbGraphicsView(self.splitter)
        self.thumb_graphicsview.setMinimumSize(QtCore.QSize(150, 0))
        self.thumb_graphicsview.setMaximumSize(QtCore.QSize(600, 16777215))
        self.thumb_graphicsview.setObjectName("thumb_graphicsview")
        self.horizontalLayout.addWidget(self.splitter)

        self.retranslateUi(librarywidget)
        self.tabWidget_2.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(librarywidget)

    def retranslateUi(self, librarywidget):
        _translate = QtCore.QCoreApplication.translate
        librarywidget.setWindowTitle(_translate("librarywidget", "Form"))
        self.tabWidget_2.setTabText(self.tabWidget_2.indexOf(self.tab_3), _translate("librarywidget", "Tab 1"))
        self.tabWidget_2.setTabText(self.tabWidget_2.indexOf(self.tab_4), _translate("librarywidget", "Tab 2"))
from docgraphicsview import DocGraphicsView
from thumbgraphicsview import ThumbGraphicsView




from resources.ui_library import Ui_librarywidget
 
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets

from popplerqt5 import Poppler

from utils import debug
from pdfworker import PdfReader

import os

class LibraryView(QtWidgets.QWidget, Ui_librarywidget):
    # renderRequested = QtCore.pyqtSignal(int, float)
    # loadFileRequested = QtCore.pyqtSignal(str, str)
    # findTextRequested = QtCore.pyqtSignal(str, int, bool)
    readAnnotationRequested = QtCore.pyqtSignal(str)
    # readOutlineRequested = QtCore.pyqtSignal(Poppler.Document, QtGui.QStandardItemModel)
    # findTextRequested = QtCore.pyqtSignal(str, int, bool)
    # pagePositionChanged = QtCore.pyqtSignal(int, int)
    # showStatusRequested = QtCore.pyqtSignal(str)

    def __init__(self, parent, screen_dpi):
        super(LibraryView, self).__init__(parent) # Call the inherited classes __init__ method
        self.setupUi(self)

        self.screen_dpi = screen_dpi

        self.splitter_doc.setSizes([1, 0]) # set relative widths of its children
        self.splitter_doc.setCollapsible(0, False) 

        self.preview_graphicsview.viewportChanged.connect(self.onDocViewportChanged)
        self.preview_graphicsview_2.viewportChanged.connect(self.onDocViewportChanged)

        self.preview_graphicsview.loadFinished.connect(self.onDoc1LoadFinished)
        self.thumb_graphicsview.loadFinished.connect(self.onThumbLoadFinished)
        self.thumb_graphicsview.pageRelocationRequest.connect(self.onPageRelocationRequest)

        # self.repos = ["/home/yhu/下载/", "/home/yhu/tmp/"]

        self.fileview_model = QtGui.QStandardItemModel(self)
        self.fileview.setModel(self.fileview_model)
        self.fileview.clicked.connect(self.onFileClick)

        self.fileview.setAlternatingRowColors(True)
        self.annotview.setAlternatingRowColors(True)

        self.annotview_model = QtGui.QStandardItemModel()
        self.annotview.setModel(self.annotview_model)
        # self.annotationView.resizeRequested.connect(self.onAnnotaionViewResized)

        # # Create separate thread for reading pdf metainformations
        # self.pdf_reader_thread = QtCore.QThread(self)
        # self.pdf_reader_obj = PdfReader()
        # self.pdf_reader_obj.moveToThread(self.pdf_reader_thread) # this must be moved before connecting signals
        # # self.renderRequested.connect(self.pdf_reader_obj.render)
        # self.readAnnotationRequested.connect(self.pdf_reader_obj.readAnnotation)
        # # self.readOutlineRequested.connect(self.pdf_reader_obj.readOutline)
        # # self.findTextRequested.connect(self.pdf_reader_obj.findText)
        # self.pdf_reader_obj.annotationFound.connect(self.onAnnotationFound)
        # # self.pdf_reader_obj.outlineFound.connect(self.onOutlineFound)
        # # self.pdf_reader_obj.textFound.connect(self.onTextFound)
        # self.pdf_reader_thread.start()

        # dirname = self.repos[0]
        # self.setFileList(dirname)
        self.current_dir = None

    def setFileList(self, dirname):
        # remove all file list first
        self.fileview_model.removeRows(0, self.fileview_model.rowCount())

        self.current_dir = dirname
        self.fileview_model.setHorizontalHeaderLabels([self.current_dir])

        parent_file_item = self.fileview_model.invisibleRootItem()

        pdf_filenames = os.listdir(dirname)
        for name in pdf_filenames:
            path = dirname + os.sep + name
            if os.path.isfile(path) and path.endswith('.pdf'):
                item = QtGui.QStandardItem(name)
                item.setData(name, QtCore.Qt.UserRole + 1)
                parent_file_item.appendRow(item)

    def setDocument(self, filename, screep_dpi):
        currentDir = os.path.dirname(filename)
        self.setFileList(currentDir)
        # 
        self.filename = filename
        self.screen_dpi = screep_dpi
        self.thumb_graphicsview.setDocument(self.filename, self.screen_dpi)

    def onFileClick(self, m_index):
        # remove all annotatoin first
        self.annotview_model.removeRows(0, self.annotview_model.rowCount())

        self.filename = self.current_dir + os.sep + self.fileview_model.data(m_index, QtCore.Qt.UserRole + 1)
        print(self.filename)

        self.thumb_graphicsview.setDocument(self.filename, self.screen_dpi)

    def onDoc1LoadFinished(self):
        # debug("onDoc1LoadFinished")
        # self.thumb_graphicsview.setDocument(self.filename, self.screen_dpi)
        pass

    def onThumbLoadFinished(self):
        # debug("onThumbLoadFinished")
        self.thumb_graphicsview.setColumnNumber(4)
        self.preview_graphicsview.setDocument(self.filename, self.screen_dpi)

    def onDocViewportChanged(self, filename, page_counts, visible_regions):
        # debug("Doc viewport changed: ", visible_regions)
        self.thumb_graphicsview.highlightVisibleMasks(filename, visible_regions)

    def onPageRelocationRequest(self, page_no, x_ratio, y_ratio):
        # debug("onPageRelocationRequest: <%d> (%.2f, %.2f)" % (page_no, x_ratio, y_ratio))
        self.preview_graphicsview.centerOnPage(page_no, x_ratio, y_ratio)

    def onAnnotationFound(self, filename, annotList):
        # the view has changed, too late
        debug(self.filename, filename)

        parent_item = self.annotview_model.invisibleRootItem()

        for pg_no in range(len(annotList)):
            for annot in annotList[pg_no]:
                image = None
                boundary = annot['boundary']
                contents = annot['contents']
                modificationDate = annot['modificationDate']
                item = QtGui.QStandardItem()
                item.setData(pg_no, QtCore.Qt.UserRole + 1)
                item.setData(boundary, QtCore.Qt.UserRole + 2)
                item.setData(contents, QtCore.Qt.UserRole + 3)
                item.setData(modificationDate, QtCore.Qt.UserRole + 4)
                # 
                timeStr = ("<%s>" % modificationDate.toString())
                if annot['type'] == 'highlight':
                    text = annot['text']
                    item.setText(text)
                elif annot['type'] == 'geom':
                    image = annot['image']
                    image = image.scaled(30, 30)
                    item.setData(QtGui.QPixmap.fromImage(image), QtCore.Qt.DecorationRole)
                    item.setData(image, QtCore.Qt.UserRole + 5)
                # item.setTextAlignment(QtCore.Qt.AlignLeft)

                dateItem = QtGui.QStandardItem()
                dateItem.setText(timeStr)
                dateItem.setTextAlignment(QtCore.Qt.AlignRight)

                parent_item.appendRow([item, dateItem])

    def goNextPage(self):
        self.preview_graphicsview.goNextPage()

    def goPrevPage(self):
        self.preview_graphicsview.goPrevPage()

    def gotoPage(self, pg_no):
        self.preview_graphicsview.gotoPage(pg_no)

    def setColumnNumber(self, colNum):
        self.preview_graphicsview.setColumnNumber(colNum)

    def setPrecedingEmptyPage(self, emptyNum):
        self.preview_graphicsview.setPrecedingEmptyPage(emptyNum)

    def zoomIn(self):
        self.preview_graphicsview.zoomIn()

    def zoomOut(self):
        self.preview_graphicsview.zoomOut()

    def zoomFitWidth(self):
        self.preview_graphicsview.zoomFitWidth()

    def closeEvent(self, ev):
        debug('closeEvent in LibraryView')
        self.preview_graphicsview.close()
        self.preview_graphicsview_2.close()
        self.thumb_graphicsview.close()
        
        # #  wait for the reader thread to exit
        # loop = QtCore.QEventLoop()
        # self.pdf_reader_thread.finished.connect(loop.quit)
        # self.pdf_reader_thread.quit()
        # loop.exec_()
        # debug('pdf_reader_thread exit')


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './kuafu/resources/ui_annotation_item.ui'
#
# Created by: PyQt5 UI code generator 5.15.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_annotationItemWidget(object):
    def setupUi(self, annotationItemWidget):
        annotationItemWidget.setObjectName("annotationItemWidget")
        annotationItemWidget.resize(451, 363)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(annotationItemWidget.sizePolicy().hasHeightForWidth())
        annotationItemWidget.setSizePolicy(sizePolicy)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(annotationItemWidget)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label_indicator = QtWidgets.QLabel(annotationItemWidget)
        self.label_indicator.setMinimumSize(QtCore.QSize(5, 0))
        self.label_indicator.setMaximumSize(QtCore.QSize(5, 16777215))
        self.label_indicator.setObjectName("label_indicator")
        self.horizontalLayout_3.addWidget(self.label_indicator)
        self.groupBox = QtWidgets.QGroupBox(annotationItemWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox.sizePolicy().hasHeightForWidth())
        self.groupBox.setSizePolicy(sizePolicy)
        self.groupBox.setTitle("")
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label_date = QtWidgets.QLabel(self.groupBox)
        self.label_date.setMaximumSize(QtCore.QSize(16777215, 20))
        self.label_date.setObjectName("label_date")
        self.horizontalLayout.addWidget(self.label_date)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.label_author = QtWidgets.QLabel(self.groupBox)
        self.label_author.setMaximumSize(QtCore.QSize(16777215, 20))
        self.label_author.setObjectName("label_author")
        self.horizontalLayout.addWidget(self.label_author)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.line = QtWidgets.QFrame(self.groupBox)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.verticalLayout.addWidget(self.line)
        self.label_title = QtWidgets.QLabel(self.groupBox)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_title.sizePolicy().hasHeightForWidth())
        self.label_title.setSizePolicy(sizePolicy)
        self.label_title.setObjectName("label_title")
        self.verticalLayout.addWidget(self.label_title)
        self.edit_comments = QtWidgets.QLineEdit(self.groupBox)
        self.edit_comments.setObjectName("edit_comments")
        self.verticalLayout.addWidget(self.edit_comments)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem1)
        self.pushButton_more = QtWidgets.QPushButton(self.groupBox)
        self.pushButton_more.setMaximumSize(QtCore.QSize(30, 30))
        self.pushButton_more.setObjectName("pushButton_more")
        self.horizontalLayout_2.addWidget(self.pushButton_more)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_3.addWidget(self.groupBox)

        self.retranslateUi(annotationItemWidget)
        QtCore.QMetaObject.connectSlotsByName(annotationItemWidget)

    def retranslateUi(self, annotationItemWidget):
        _translate = QtCore.QCoreApplication.translate
        annotationItemWidget.setWindowTitle(_translate("annotationItemWidget", "Form"))
        self.label_indicator.setText(_translate("annotationItemWidget", "TextLabel"))
        self.label_date.setText(_translate("annotationItemWidget", "TextLabel"))
        self.label_author.setText(_translate("annotationItemWidget", "TextLabel"))
        self.label_title.setText(_translate("annotationItemWidget", "TextLabel"))
        self.pushButton_more.setText(_translate("annotationItemWidget", "..."))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './kuafu/resources/ui_document.ui'
#
# Created by: PyQt5 UI code generator 5.15.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_doc_view(object):
    def setupUi(self, doc_view):
        doc_view.setObjectName("doc_view")
        doc_view.resize(843, 432)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(doc_view.sizePolicy().hasHeightForWidth())
        doc_view.setSizePolicy(sizePolicy)
        self.horizontalLayout = QtWidgets.QHBoxLayout(doc_view)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.splitter_main = QtWidgets.QSplitter(doc_view)
        self.splitter_main.setOrientation(QtCore.Qt.Horizontal)
        self.splitter_main.setObjectName("splitter_main")
        self.tabWidget1 = QtWidgets.QTabWidget(self.splitter_main)
        self.tabWidget1.setMaximumSize(QtCore.QSize(400, 16777215))
        self.tabWidget1.setObjectName("tabWidget1")
        self.outline = QtWidgets.QWidget()
        self.outline.setObjectName("outline")
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout(self.outline)
        self.horizontalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.treeView_outline = QtWidgets.QTreeView(self.outline)
        self.treeView_outline.setObjectName("treeView_outline")
        self.horizontalLayout_5.addWidget(self.treeView_outline)
        self.tabWidget1.addTab(self.outline, "")
        self.thumbs = QtWidgets.QWidget()
        self.thumbs.setObjectName("thumbs")
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout(self.thumbs)
        self.horizontalLayout_6.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.graphicsView_thumbs = DocGraphicsView(self.thumbs)
        self.graphicsView_thumbs.setObjectName("graphicsView_thumbs")
        self.horizontalLayout_6.addWidget(self.graphicsView_thumbs)
        self.tabWidget1.addTab(self.thumbs, "")
        self.splitter_doc = QtWidgets.QSplitter(self.splitter_main)
        self.splitter_doc.setOrientation(QtCore.Qt.Vertical)
        self.splitter_doc.setObjectName("splitter_doc")
        self.graphicsView_doc1 = DocGraphicsView(self.splitter_doc)
        self.graphicsView_doc1.setObjectName("graphicsView_doc1")
        self.graphicsView_doc2 = DocGraphicsView(self.splitter_doc)
        self.graphicsView_doc2.setObjectName("graphicsView_doc2")
        self.tabWidget2 = QtWidgets.QTabWidget(self.splitter_main)
        self.tabWidget2.setMinimumSize(QtCore.QSize(0, 0))
        self.tabWidget2.setMaximumSize(QtCore.QSize(400, 16777215))
        self.tabWidget2.setTabBarAutoHide(True)
        self.tabWidget2.setObjectName("tabWidget2")
        self.annotations = QtWidgets.QWidget()
        self.annotations.setObjectName("annotations")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.annotations)
        self.verticalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.annotScrollArea = CustomScrollArea(self.annotations)
        self.annotScrollArea.setWidgetResizable(True)
        self.annotScrollArea.setObjectName("annotScrollArea")
        self.annotScrollAreaContents = QtWidgets.QWidget()
        self.annotScrollAreaContents.setGeometry(QtCore.QRect(0, 0, 122, 428))
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.annotScrollAreaContents.sizePolicy().hasHeightForWidth())
        self.annotScrollAreaContents.setSizePolicy(sizePolicy)
        self.annotScrollAreaContents.setObjectName("annotScrollAreaContents")
        self.annotScrollAreaContents_Layout = QtWidgets.QHBoxLayout(self.annotScrollAreaContents)
        self.annotScrollAreaContents_Layout.setContentsMargins(0, 0, 0, 0)
        self.annotScrollAreaContents_Layout.setObjectName("annotScrollAreaContents_Layout")
        self.annotScrollArea.setWidget(self.annotScrollAreaContents)
        self.verticalLayout_4.addWidget(self.annotScrollArea)
        self.tabWidget2.addTab(self.annotations, "")
        self.horizontalLayout.addWidget(self.splitter_main)

        self.retranslateUi(doc_view)
        self.tabWidget1.setCurrentIndex(1)
        QtCore.QMetaObject.connectSlotsByName(doc_view)

    def retranslateUi(self, doc_view):
        _translate = QtCore.QCoreApplication.translate
        doc_view.setWindowTitle(_translate("doc_view", "Form"))
        self.tabWidget1.setTabText(self.tabWidget1.indexOf(self.outline), _translate("doc_view", "Outline"))
        self.tabWidget1.setTabText(self.tabWidget1.indexOf(self.thumbs), _translate("doc_view", "Thumbs"))
        self.tabWidget2.setTabText(self.tabWidget2.indexOf(self.annotations), _translate("doc_view", "Annotations"))
from customscrollarea import CustomScrollArea
from docgraphicsview import DocGraphicsView




# -*- coding: utf-8 -*-
import time
from email.utils import parsedate_tz, mktime_tz

from PyQt5 import QtCore
from PyQt5.QtGui import QIcon, QIntValidator
from PyQt5.QtWidgets import ( QDialog, QDialogButtonBox, QGridLayout, QLineEdit, QSpinBox,
    QLabel, QVBoxLayout, QPushButton, QTableWidget, QTableWidgetItem, QHeaderView
)

class ExportToImageDialog(QDialog):
    def __init__(self, page_no, total_pages, parent):
        QDialog.__init__(self, parent)
        self.setWindowTitle('Export Page to Image')
        self.resize(300, 140)
        layout = QGridLayout(self)
        dpiLabel = QLabel('DPI :', self)
        self.dpiEdit = QLineEdit("300", self)
        self.dpiEdit.setAlignment(QtCore.Qt.AlignHCenter)
        self.dpiEdit.setValidator(QIntValidator(75, 1200, self))
        pageNoLabel = QLabel('Page No. :', self)
        self.pageNoSpin = QSpinBox(self)
        self.pageNoSpin.setAlignment(QtCore.Qt.AlignHCenter)
        self.buttonBox = QDialogButtonBox(self)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Save|QDialogButtonBox.Cancel)
        layout.addWidget(dpiLabel, 0,0,1,1)
        layout.addWidget(self.dpiEdit, 0,1,1,1)
        layout.addWidget(pageNoLabel, 1,0,1,1)
        layout.addWidget(self.pageNoSpin, 1,1,1,1)
        layout.addWidget(self.buttonBox, 2, 0, 1, 2)

        # set values
        self.pageNoSpin.setRange(1, total_pages)
        self.pageNoSpin.setValue(page_no)
        # connect signals
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)


class DocInfoDialog(QDialog):
    def __init__(self, parent):
        QDialog.__init__(self, parent)
        self.resize(560, 320)
        self.tableWidget = QTableWidget(0, 2, self)
        vLayout = QVBoxLayout(self)
        vLayout.addWidget(self.tableWidget)
        self.tableWidget.setAlternatingRowColors(True)
        closeBtn = QPushButton(QIcon(':/quit.png'), "Close", self)
        closeBtn.setMaximumWidth(120)
        vLayout.addWidget(closeBtn, 0, QtCore.Qt.AlignRight)
        closeBtn.clicked.connect(self.accept)
        self.tableWidget.horizontalHeader().setDefaultSectionSize(150)
        self.tableWidget.horizontalHeader().setSectionResizeMode(1, QHeaderView.Stretch)
        self.tableWidget.horizontalHeader().setVisible(False)
        self.tableWidget.verticalHeader().setVisible(False)

    def setInfo(self, info_keys, values):
        for i in range(len(info_keys)):
            if info_keys[i] in ['ModDate', 'CreationDate']:
                values[i] = parsePdfTime(values[i])
            self.tableWidget.insertRow(i)
            self.tableWidget.setItem(i,0, QTableWidgetItem(info_keys[i]))
            self.tableWidget.setItem(i,1, QTableWidgetItem(values[i]))

# Takes D:20130501200439+01'00' like format and returns a local timezone based format
# In some pdfs date does not start with D:
def parsePdfTime(t):
    t = t.replace('D:', '')
    try:
        ts = time.strptime(t[:14], "%Y%m%d%H%M%S")
    except:
        return t
    rfc2822 = time.strftime("%a, %d %b %Y %H:%M:%S "+t[-5:], ts) # to rfc2822 time format
    py_time = mktime_tz(parsedate_tz(rfc2822))
    return time.strftime("%d %b %Y at %I:%M:%S %p", time.localtime(py_time))





from PyQt5 import QtWidgets
from PyQt5 import QtCore
from PyQt5 import QtGui

from basedocgraphicsview import BaseDocGraphicsView
from utils import debug

class ThumbGraphicsView(BaseDocGraphicsView):
    pageRelocationRequest = QtCore.pyqtSignal(int, float, float)

    def __init__(self, parent, render_num=4):
        super(ThumbGraphicsView, self).__init__(parent, render_num)

        self.scene.setBackgroundBrush(QtGui.QBrush(QtCore.Qt.white)) # set background

        self.isMouseOver = False

    def highlightVisibleMasks(self, filename, visible_regions):
        # the items in visible_regions are in normalized coordinates
        # 
        if filename != self.current_filename:
            return
        if not self.load_finished_flag:
            return
        # 
        # remove current highlighted first
        for pg_no in self.current_highlighted_pages:
            self.page_items[pg_no].setMask(0, 0, 0, 0)
        self.current_highlighted_pages = []
        # remove border of the page that is marked as the current
        self.page_items[self.pageMarkedAsCurrent].setBorderHighlight(False)

        currentPageIdx = next(iter(visible_regions)) # first key as the current
        for pg_no in visible_regions:
            vRect = visible_regions[pg_no]
            self.initializePage(pg_no)
            if pg_no == currentPageIdx:
                self.page_items[pg_no].setBorderHighlight(True)
                self.pageMarkedAsCurrent = pg_no
            self.page_items[pg_no].setMask(vRect[0], vRect[1], vRect[2], vRect[3])
            if not self.isMouseOver:
                self.centerOn(self.page_items[pg_no])
            # 
            self.current_highlighted_pages.append(pg_no)

    def mousePressEvent(self, ev):
        # debug('mousePressEvent in ThumbGraphicsView')
        if not self.load_finished_flag:
            return
        page_no, x_ratio, y_ratio = self.getPageByPos(ev.pos().x(), ev.pos().y())
        self.pageRelocationRequest.emit(page_no, x_ratio, y_ratio)
        return super().mousePressEvent(ev)

    def enterEvent(self, ev):
        # debug('enterEvent in ThumbGraphicsView')
        self.isMouseOver = True
        return super().enterEvent(ev)
    
    def leaveEvent(self, ev):
        # debug('leaveEvent in ThumbGraphicsView')
        self.isMouseOver = False
        return super().leaveEvent(ev)



from PyQt5 import QtWidgets
from PyQt5 import QtCore
from PyQt5 import QtGui
# from PyQt5 import QtOpenGL

# import fitz #PyMuPDF
from popplerqt5 import Poppler

from pdfworker import PdfRender
from multiprocessing import Queue
from page import PageGraphicsItem

from utils import debug

import math
import time
import numpy as np

class BaseDocGraphicsView(QtWidgets.QGraphicsView):
    loadFinished = QtCore.pyqtSignal()
    viewportChanged = QtCore.pyqtSignal(str, int, dict)
    zoomRatioChanged = QtCore.pyqtSignal(float)
    viewColumnChanged = QtCore.pyqtSignal(int)
    emptyLeadingPageChanged = QtCore.pyqtSignal(int)

    def __init__(self, parent, render_num=4):
        super(BaseDocGraphicsView, self).__init__(parent)

        # self.setViewport(QtOpenGL.QGLWidget()) # opengl

        self.scene = QtWidgets.QGraphicsScene(self)

        self.current_cursor_x = 0
        self.current_cursor_y = 0

        self.setScene(self.scene)

        self.page_items = [] # instances of PageGraphicsItem
        self.load_finished_flag = False

        self.current_filename = None

        self.page_counts = 0
        self.pages_size_inch = [] # (width, height) in inch, invarient
        self.current_pages_size_pix = [] # varient depends on the view
        self.current_pages_rect = [] # one for flag and 4 for scene coordinates
        self.pageMarkedAsCurrent = 0

        self.current_visible_regions = {}
        self.view_column_count = 1
        self.leading_empty_pages = 0

        self.render_num = render_num
        self.render_list = []
        self.rendered_info = {}
        self.current_rendering_dpi = []
        self.current_highlighted_pages = []

        for i in range(self.render_num):
            tmpRender = PdfRender(Queue(), Queue())
            tmpRender.start()
            self.render_list.append(tmpRender)

        self.screen_dpi = 0

        self.zoom_levels = [0.12, 0.25, 0.33, 0.50, 0.66, 0.75, 1.0, 1.25, 1.5, 2.0, 4.0, 8.0, 16.0]
        self.current_zoom_index = -1
        self.fitwidth_flag = True

        self.horispacing = 3
        self.vertspacing = 5

        self.horizontalScrollBar().valueChanged.connect(self.onScrollValueChanged)
        self.verticalScrollBar().valueChanged.connect(self.onScrollValueChanged)

        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)

        # self.setRenderHint(QtGui.QPainter.Antialiasing)
        self.setViewportUpdateMode(QtWidgets.QGraphicsView.FullViewportUpdate)
        # self.setViewportUpdateMode(QtWidgets.QGraphicsView.SmartViewportUpdate)
        # self.setViewportUpdateMode(QtWidgets.QGraphicsView.MinimalViewportUpdate)
        self.setDragMode(QtWidgets.QGraphicsView.ScrollHandDrag)

        self.setMouseTracking(True)

        self.setTransform(QtGui.QTransform()) # set identity matrix for transformation

        # handler for croll value changed
        self.scrollValueChanged_flag = False
        self.scrollValueChanged_timer = QtCore.QTimer(self)
        self.scrollValueChanged_timer.timeout.connect(self.scrollValueChangedHandler)
        self.scrollValueChanged_timer.start(20)

        # resize handler
        self.resized_flag = False
        self.resize_timer = QtCore.QTimer(self)
        self.resize_timer.timeout.connect(self.resizeHandler)
        self.resize_timer.start(200)

        # read the queue periodically
        self.queue_timer = QtCore.QTimer(self)
        self.queue_timer.timeout.connect(self.retrieveQueueResults)
        self.queue_timer.start(20)

    def clear(self):
        self.scene.clear()
        self.page_items = []
        self.pages_size_inch = [] # (width, height) in inch, invarient
        self.current_pages_size_pix = [] # varient depends on the view
        self.current_pages_rect = []
        self.pageMarkedAsCurrent = 0
        self.current_visible_regions = {}
        self.current_rendering_dpi = []
        self.rendered_info = {}
        self.current_highlighted_pages = []
        self.load_finished_flag = False

    def setDocument(self, filename, screen_dpi):
        self.clear()

        self.current_filename = filename
        self.screen_dpi = screen_dpi

        # reSET for all render processes
        for rd in self.render_list:
            rd.commandQ.put(['SET', [self.current_filename]])

        # query page size using the first worker
        self.render_list[0].commandQ.put(['PAGESIZES', [None]])

    def onPageSizesReceived(self, pages_size_inch):
        debug("%d Page Sizes Received" % len(pages_size_inch))

        # time_0 = time.time()
        self.page_counts = len(pages_size_inch)
        self.pages_size_inch = pages_size_inch

        for i in range(self.page_counts):
            pageItem = PageGraphicsItem(i)
            self.scene.addItem(pageItem)
            self.page_items.append(pageItem)

        self.fitwidth_flag = True
        # time_a = time.time()
        # debug('A', time_a - time_0)

        self.computePagesDPI()
        # time_b = time.time()
        # debug('B', time_b - time_a)

        self.__rearrangePages()
        # time_c = time.time()
        # debug('C', time_c - time_b)

        self.setColumnNumber(min(self.view_column_count, self.page_counts))
        self.setPrecedingEmptyPage(0)
        # 
        self.onViewportChanged()
        # 
        self.load_finished_flag = True
        self.loadFinished.emit()

    def onViewportChanged(self):
        self.renderCurrentVisiblePages()
        # 
        # prepare normalized visible regions
        if len(self.current_visible_regions) == 0:
            return
        vRegions = {}
        for pg_no in self.current_visible_regions:
            _, x, y, w, h = self.current_pages_rect[pg_no]
            rect = self.current_visible_regions[pg_no]
            normalized_rect = [rect.x() / w, rect.y()/ h, rect.width() / w, rect.height() / h]
            vRegions[pg_no] = normalized_rect
        self.viewportChanged.emit(self.current_filename, self.page_counts, vRegions)

    def computePagesDPI(self):
        self.current_rendering_dpi = []
        self.current_pages_size_pix =[]
        if self.page_counts == 0:
            return

        if self.fitwidth_flag:
            # all pages will have the same width, but may different height
            viewport = self.viewport()
            viewWidth = viewport.width()
            viewHeight = viewport.height()
            
            viewWidth -= (self.horispacing * (self.view_column_count + 1))
            viewWidth /= self.view_column_count
            
            avg_zLevel = 0 # compute the nearest zoom level for fitting width mode
            for i in range(self.page_counts):
                dpi = viewWidth / self.pages_size_inch[i][0]
                height = self.pages_size_inch[i][1] * dpi
                self.current_rendering_dpi.append(dpi)
                self.current_pages_size_pix.append([int(viewWidth), int(height)])
                # 
                avg_zLevel += (dpi / self.screen_dpi)
            # 
            avg_zLevel /= self.page_counts
            candi_zLevels = np.array(self.zoom_levels)
            diff = np.abs(candi_zLevels - avg_zLevel)
            self.current_zoom_index = np.argmin(diff)
        else:
            zLevel = self.zoom_levels[self.current_zoom_index]
            dpi = self.screen_dpi * zLevel # every page will have the same dpi
            for i in range(self.page_counts):
                width = self.pages_size_inch[i][0] * dpi
                height = self.pages_size_inch[i][1] * dpi
                self.current_rendering_dpi.append(dpi)
                self.current_pages_size_pix.append([int(width), int(height)])

    def getVisibleRegions(self):
        visRect = self.viewport().rect() # visible area
        visRect = self.mapToScene(visRect).boundingRect() # change to scene coordinates
        # 
        regions = {}
        for pg_no in range(self.page_counts):
            flag, x, y, w, h = self.current_pages_rect[pg_no]
            pageRect = QtCore.QRectF(x, y, w, h)
            intsec = visRect.intersected(pageRect)
            # change to item coordinate
            intsec = QtCore.QRectF(
                intsec.x() - x, intsec.y() - y, 
                intsec.width(), intsec.height()
                )
            if intsec.isEmpty():
                continue
            regions[pg_no] = intsec

        self.current_visible_regions = regions

    def renderRequest(self, page_no, dpi, roi, render_idx, visible_regions):
        self.render_list[render_idx].commandQ.put(['RENDER', [page_no, dpi, roi, visible_regions]])

    def initializePage(self, page_no):
        if not self.current_pages_rect[page_no][0]:
            _, x, y, w, h = self.current_pages_rect[page_no]
            self.page_items[page_no].initialize(x, y, w, h)
            self.current_pages_rect[page_no][0] = True

    def renderCurrentVisiblePages(self):
        self.getVisibleRegions()

        if len(self.current_visible_regions) == 0:
            return

        # debug("Visible Regions: ", self.current_visible_regions)
        
        # update the cached images for visible pages first (for smoothing transitions)
        for page_no in self.current_visible_regions:
            # initilize page first
            self.initializePage(page_no)
            self.page_items[page_no].updateTransientItems(self.current_visible_regions[page_no])

        for page_no in self.current_visible_regions:
            history_dpi = 0
            history_roi_list = []
            if page_no in self.rendered_info:
                history_dpi, history_roi_list = self.rendered_info[page_no]
                
            # split roi to small patches
            roi_raw = self.current_visible_regions[page_no]
            patch_positions, patches = self.page_items[page_no].get_roi_patches(roi_raw)
            patch_col_num = self.page_items[page_no].patch_col_num

            dpi = self.current_rendering_dpi[page_no]
            for i in range(len(patches)):
                pRow, pCol = patch_positions[i]
                pIdx = pRow * patch_col_num + pCol
                roi = patches[i]

                # already cached, no need to render
                if dpi == history_dpi and roi in history_roi_list:
                    continue

                # assign render index according to the page number and the patch position
                prefixNum = page_no % self.render_num
                render_idx = (prefixNum + pIdx) % self.render_num

                self.renderRequest(page_no, dpi, roi, render_idx, self.current_visible_regions)

                # debug("<- Render %d Requested : <page:%d> <dpi:%.2f> <roi_raw: %.1f %.1f %.1f %.1f> <roi: %.1f %.1f %.1f %.1f>" % (
                #     render_idx, page_no, dpi, 
                #     roi_raw.left(), roi_raw.top(), roi_raw.width(), roi_raw.height(), 
                #     roi.left(), roi.top(), roi.width(), roi.height()
                #     ))

    def handleSingleRenderedImage(self, render_idx, filename, page_no, dpi, roi, img_byte_array):
        # debug("-> Rendering %d Completed : <page:%d> <dpi:%.2f> <roi: %.1f %.1f %.1f %.1f>" % (
        #     render_idx, page_no, dpi, roi.left(), roi.top(), roi.width(), roi.height()
        # ))

        # the doc has changed, too late
        if filename != self.current_filename:
            # debug("file name changed: %s -> %s. skipping" % (filename, self.current_filename))
            return

        # if page_no not in self.current_visible_regions:
        #     debug("become unvisible: %d. skipping" % page_no)
        #     return
        
        if page_no in self.rendered_info \
           and self.rendered_info[page_no][0] == dpi \
           and roi in self.rendered_info[page_no][1]:
            # debug("duplicated rendering. skipping")
            return

        if len(self.current_rendering_dpi) == 0:
            return
        # too late, the current rendering dpi is already changed
        if dpi != self.current_rendering_dpi[page_no]:
            # debug("unmatched DPI: %.2f -> %.2f. skipping" % (dpi, self.current_rendering_dpi[page_no]))
            return
            
        # QByteArray to QImage
        img_buffer = QtCore.QBuffer(img_byte_array)
        img_buffer.open(QtCore.QIODevice.ReadOnly)
        image = QtGui.QImageReader(img_buffer).read()

        # crop to container's size
        # containerSize = self.page_items[page_no].rect()
        # debug("dim: (%d x %d) -> [%d x %d]" % (image.width(), image.height(), containerSize.width(), containerSize.height()))
        # roi = QtCore.QRect(1,1,containerSize.width()-2, containerSize.height()-2)
        # image = image.copy(roi)

        self.page_items[page_no].addPixmap(QtGui.QPixmap.fromImage(image), roi.x(), roi.y())
        
        if page_no in self.rendered_info:
            dpi0, roi_list = self.rendered_info[page_no]
            assert(dpi == dpi0)
            roi_list.append(roi)
        else:
            self.rendered_info[page_no] = [dpi, [roi]]

        # # Request to render next page
        # if self.current_page <= page_no < (self.current_page + self.max_preload - 2):
        #     if (page_no+2 not in self.rendered_pages) and (page_no+2 <= self.pages_count):
        #       self.renderRequested.emit(page_no+2, self.pages[page_no+1].dpi)
        # Replace old rendered pages with blank image

        #if len(self.rendered_info) > 10:
            #self.getVisibleRegions() # update visible regions again
            #firstKey = next(iter(self.rendered_info))
            #if self.current_visible_regions is None or firstKey not in self.current_visible_regions:
                #self.rendered_info.pop(firstKey)
                #debug("Clear Page :", firstKey)
                #self.page_items[firstKey].clear()

        # debug("Rendered Images: ", self.rendered_pages)

    def retrieveQueueResults(self):
        for rd_idx in range(self.render_num):
            rd = self.render_list[rd_idx]
            resultsQ = rd.resultsQ
            # collect all results
            size = resultsQ.qsize()
            for i in range(size):
                item = resultsQ.get() # will be blocked until when some data are avaliable
                if item[0] == 'PAGESIZES_RES':
                    _, filename, pages_size_inch = item
                    if filename != self.current_filename: # the doc has changed, too late
                        continue
                    self.onPageSizesReceived(pages_size_inch)
                elif item[0] == 'RENDER_RES':
                    _, filename, page_no, dpi, roi, img_byte_array = item
                    self.handleSingleRenderedImage(rd_idx, filename, page_no, dpi, roi, img_byte_array)

    def __rearrangePages(self):
        if len(self.current_pages_size_pix) == 0:
            return

        pages_width_pix = np.array(self.current_pages_size_pix)[:, 0]
        pages_height_pix = np.array(self.current_pages_size_pix)[:, 1]

        rowNum = math.ceil((self.page_counts + self.leading_empty_pages) / self.view_column_count)
        pages_width_pix = np.concatenate((np.zeros((self.leading_empty_pages)), pages_width_pix))
        pages_height_pix = np.concatenate((np.zeros((self.leading_empty_pages)), pages_height_pix))

        pad_num = (self.page_counts + self.leading_empty_pages) % self.view_column_count
        if pad_num > 0:
            pad_num = self.view_column_count - pad_num
            pages_width_pix = np.concatenate((pages_width_pix, np.zeros((pad_num))))
            pages_height_pix = np.concatenate((pages_height_pix, np.zeros((pad_num))))

        pages_width_pix = pages_width_pix.reshape((rowNum, self.view_column_count))
        pages_height_pix = pages_height_pix.reshape((rowNum, self.view_column_count))
        rowHeights = pages_height_pix.max(axis=1)
        colWidths = pages_width_pix.max(axis=0)

        rowCumHeights = np.cumsum(rowHeights)
        colCumWidths = np.cumsum(colWidths)

        self.current_pages_rect = []
        for i in range(self.page_counts):
            row = int((i + self.leading_empty_pages) / self.view_column_count)
            col = int((i + self.leading_empty_pages) % self.view_column_count)

            startx = colCumWidths[col - 1] if col > 0 else 0
            starty = rowCumHeights[row - 1] if row > 0 else 0

            # add spacing 
            startx += (self.horispacing * (col + 1))
            starty += (self.vertspacing * (row + 1))

            # center it
            startx += (colWidths[col] - pages_width_pix[row][col]) / 2
            starty += (rowHeights[row] - pages_height_pix[row][col]) / 2

            # save page rect, False means uninitialized
            self.current_pages_rect.append([False, startx, starty, pages_width_pix[row][col], pages_height_pix[row][col]])
            
            # make all pages invisible first
            self.page_items[i].setVisible(False)

        sceneWidthFix = colWidths.sum() + (self.view_column_count + 1) * self.horispacing
        sceneHeightFix = rowHeights.sum() + (rowNum + 1) * self.vertspacing
        self.scene.setSceneRect(0, 0, sceneWidthFix, sceneHeightFix)
        
        # after rearrangement, the cached rendering information is useless
        # that means, everything should be re-rendered
        self.rendered_info = {}

    def centerOnPage(self, page_no, x_ratio, y_ratio):
        _, x, y, w, h = self.current_pages_rect[page_no]
        cx = x + x_ratio * w
        cy = y + y_ratio * h
        self.centerOn(cx, cy)

    def getPageByPos(self, view_x, view_y):
        item = self.itemAt(view_x, view_y)
        if item:
            item = item.parentItem() if item.parentItem() else item
            assert(isinstance(item, PageGraphicsItem))
            page_no_cursor_at = self.page_items.index(item)
        else:
            # at empty places, find the nearest page
            scene_pos = self.mapToScene(view_x, view_y)
            page_rects = np.array(self.current_pages_rect)[:, 1:]
            pages_cx = page_rects[:, 0] + page_rects[:, 2] / 2
            pages_cy = page_rects[:, 1] + page_rects[:, 3] / 2
            distance = np.abs(pages_cx - scene_pos.x()) + np.abs(pages_cy - scene_pos.y()) # L1 distance is enough
            page_no_cursor_at = np.argmin(distance)
        # 
        _, x, y, w, h = self.current_pages_rect[page_no_cursor_at]
        cursor_in_scene = self.mapToScene(self.current_cursor_x, self.current_cursor_y)
        x_ratio = (cursor_in_scene.x() - x) / w # the relative position
        y_ratio = (cursor_in_scene.y() - y) / h

        return page_no_cursor_at, x_ratio, y_ratio

    def redrawPages(self):
        if self.page_counts == 0 or len(self.current_pages_rect) == 0:
            return

        # time_0 = time.time()

        # save previous cursor position
        page_no_cursor_at, x_ratio, y_ratio = self.getPageByPos(
            self.current_cursor_x, self.current_cursor_y
            )
        viewRect = self.viewport()
        dx_center = viewRect.width() / 2 - self.current_cursor_x
        dy_center = viewRect.height() / 2 - self.current_cursor_y

        # compute DPI
        self.computePagesDPI()
        # time_a = time.time()
        # debug("A ", time_a - time_0)

        self.__rearrangePages()
        # time_b = time.time()
        # debug("B ", time_b - time_a)

        # set viewport after rearranging (make page position under cursor unchanged)
        _, new_x, new_y, new_w, new_h = self.current_pages_rect[page_no_cursor_at]
        scene_cx = new_x + x_ratio * new_w + dx_center
        scene_cy = new_y + y_ratio * new_h + dy_center
        self.centerOn(scene_cx, scene_cy)

        self.onViewportChanged()

    def setColumnNumber(self, colNum):
        if self.page_counts > 0:
            self.view_column_count = min(self.page_counts, colNum)
            self.redrawPages()
            self.viewColumnChanged.emit(self.view_column_count)

    def setPrecedingEmptyPage(self, emptyPages):
        if self.page_counts > 0:
            if self.view_column_count > 1:
                self.leading_empty_pages = emptyPages
            else:
                self.leading_empty_pages = 0
            self.redrawPages()
            self.emptyLeadingPageChanged.emit(self.leading_empty_pages)

    def zoomIn(self):
        if self.current_zoom_index == len(self.zoom_levels) - 1:
            return
        self.current_zoom_index += 1
        self.fitwidth_flag = False
        self.redrawPages()
        self.zoomRatioChanged.emit(self.zoom_levels[self.current_zoom_index])

    def zoomOut(self):
        if self.current_zoom_index == 0:
            return
        self.current_zoom_index -= 1
        self.fitwidth_flag = False
        self.redrawPages()
        self.zoomRatioChanged.emit(self.zoom_levels[self.current_zoom_index])

    def zoomFitWidth(self):
        self.fitwidth_flag = True
        self.redrawPages()
        self.zoomRatioChanged.emit(0)

    def goPrevPage(self):
        if len(self.current_visible_regions) == 0:
            return
        curent_page_idx = next(iter(self.current_visible_regions)) # first key as the current
        self.gotoPage(curent_page_idx - 1)

    def goNextPage(self):
        if len(self.current_visible_regions) == 0:
            return
        curent_page_idx = next(iter(self.current_visible_regions)) # first key as the current
        self.gotoPage(curent_page_idx + 1)

    def gotoPage(self, pg_no):
        pass

    # def showEvent(self, ev):
    #     debug('showEvent in BaseDocGraphicsView')
    #     ev.ignore()

    def closeEvent(self, ev):
        debug('closeEvent in BaseDocGraphicsView')
        self.destroyRenders()
        debug('All renders are destroyed')

    def destroyRenders(self):
        for rd in self.render_list:
            rd.commandQ.put(['STOP', []])
        for rd in self.render_list:
            # rd.wait()
            rd.join()
        self.render_list = []
        self.render_num = 0

    def onScrollValueChanged(self):
        self.scrollValueChanged_flag = True

    def scrollValueChangedHandler(self):
        if self.scrollValueChanged_flag:
            self.onViewportChanged()
            self.scrollValueChanged_flag = False

    def resizeHandler(self):
        if self.resized_flag:
            self.redrawPages()
            self.resized_flag = False

    def resizeEvent(self, ev):
        # debug('resizeEvent in BaseDocGraphicsView')
        self.resized_flag = True
        # call the parent's handler, it will do some alignments
        return super().resizeEvent(ev)

    def mousePressEvent(self, ev):
        # debug('mousePressEvent in BaseDocGraphicsView')
        return super().mousePressEvent(ev)

    def mouseReleaseEvent(self, ev):
        # debug('mouseReleaseEvent in BaseDocGraphicsView')
        return super().mouseReleaseEvent(ev)

    def mouseMoveEvent(self, ev):
        # debug('mouseMoveEvent in BaseDocGraphicsView')
        self.current_cursor_x = ev.x()
        self.current_cursor_y = ev.y()
        # debug(self.current_cursor_x, self.current_cursor_y)
        return super().mouseMoveEvent(ev)



from PyQt5 import QtWidgets
from PyQt5 import QtCore
from PyQt5 import QtGui
from popplerqt5 import Poppler

from utils import debug

class CustomScrollArea(QtWidgets.QScrollArea):
    # rendered = QtCore.pyqtSignal(int, QtGui.QImage)
    # textFound = QtCore.pyqtSignal(int, list)

    resizeRequested = QtCore.pyqtSignal(int, int)

    def __init__(self, parent):
        super(CustomScrollArea, self).__init__(parent)
        
        # emit the resize event slowly
        self.resize_event_queue = []
        self.resize_timer = QtCore.QTimer(self)
        self.resize_timer.timeout.connect(self.resizeTrigger)
        self.resize_timer.start(200)

    def resizeTrigger(self):
        if len(self.resize_event_queue) > 0:
            self.resizeRequested.emit(self.width(), self.height())
            self.resize_event_queue = []

    def resizeEvent(self, ev):
        # debug("resizeEvent in DocScrollArea")
        # print(self.width(), self.height())
        self.resize_event_queue.append(1) # save first
        QtWidgets.QScrollArea.resizeEvent(self, ev) # call parent's event handler
        
